import ctypes
import errno
import fcntl


# Define ioctl-related functions and structures
def _IOC(dir, type, nr, size):
    IOC_NRBITS = 8
    IOC_TYPEBITS = 8
    IOC_SIZEBITS = 14
    IOC_DIRBITS = 2  # noqa

    IOC_NRSHIFT = 0
    IOC_TYPESHIFT = IOC_NRSHIFT + IOC_NRBITS
    IOC_SIZESHIFT = IOC_TYPESHIFT + IOC_TYPEBITS
    IOC_DIRSHIFT = IOC_SIZESHIFT + IOC_SIZEBITS

    return (
        (dir << IOC_DIRSHIFT)
        | (ord(type) << IOC_TYPESHIFT)
        | (nr << IOC_NRSHIFT)
        | (size << IOC_SIZESHIFT)
    )


IOC_NONE = 0
IOC_WRITE = 1
IOC_READ = 2


def _IOWR(type, nr, size):
    return _IOC(IOC_READ | IOC_WRITE, type, nr, size)


class fiemap_extent(ctypes.Structure):
    _fields_ = [
        ("fe_logical", ctypes.c_uint64),
        ("fe_physical", ctypes.c_uint64),
        ("fe_length", ctypes.c_uint64),
        ("fe_reserved64", ctypes.c_uint64 * 2),
        ("fe_flags", ctypes.c_uint32),
        ("fe_reserved", ctypes.c_uint32 * 3),
    ]


# Define the base fiemap structure without fm_extents
class fiemap_base(ctypes.Structure):
    _fields_ = [
        ("fm_start", ctypes.c_uint64),
        ("fm_length", ctypes.c_uint64),
        ("fm_flags", ctypes.c_uint32),
        ("fm_mapped_extents", ctypes.c_uint32),
        ("fm_extent_count", ctypes.c_uint32),
        ("fm_reserved", ctypes.c_uint32),
        # Do not include fm_extents here
    ]


def create_fiemap_struct(extent_count):
    class fiemap(ctypes.Structure):
        _fields_ = [
            ("fm_start", ctypes.c_uint64),
            ("fm_length", ctypes.c_uint64),
            ("fm_flags", ctypes.c_uint32),
            ("fm_mapped_extents", ctypes.c_uint32),
            ("fm_extent_count", ctypes.c_uint32),
            ("fm_reserved", ctypes.c_uint32),
            ("fm_extents", fiemap_extent * extent_count),
        ]

    return fiemap


def get_extents(fd):
    # This function retrieves all extents for a given file descriptor
    extents = []
    last_logical = 0
    # block_size = os.statvfs(fd).f_bsize

    while True:
        extent_count = 32  # Number of extents to retrieve per ioctl call
        fiemap_struct = create_fiemap_struct(extent_count)
        fm = fiemap_struct()
        fm.fm_start = last_logical
        fm.fm_length = ctypes.c_uint64(-1).value  # Equivalent to ULLONG_MAX
        fm.fm_flags = 0
        fm.fm_extent_count = extent_count

        fiemap_base_size = ctypes.sizeof(fiemap_base)
        FS_IOC_FIEMAP = _IOWR("f", 11, fiemap_base_size)

        # Perform the ioctl call
        try:
            fcntl.ioctl(fd, FS_IOC_FIEMAP, fm)
        except OSError as e:
            if e.errno == errno.ENOTTY:
                raise NotImplementedError(
                    "FIEMAP ioctl is not supported on this filesystem"
                )
            else:
                raise

        if fm.fm_mapped_extents == 0:
            break  # No more extents

        for i in range(fm.fm_mapped_extents):
            fe = fm.fm_extents[i]
            extent = {
                "logical": fe.fe_logical,
                "physical": fe.fe_physical,
                "length": fe.fe_length,
                "flags": fe.fe_flags,
            }
            extents.append(extent)
            last_logical = fe.fe_logical + fe.fe_length

        # Check if the last extent has the FIEMAP_EXTENT_LAST flag
        if fm.fm_extents[fm.fm_mapped_extents - 1].fe_flags & 0x1:  # FIEMAP_EXTENT_LAST
            break

    return extents



#!/usr/bin/env python3
import argparse

from .filemap import FileMap


def main():
    parser = argparse.ArgumentParser(description="Prints details about file fragments")
    parser.add_argument(
        "-x", "--hex", action="store_true", help="Output numbers in hexadecimal"
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Display detailed extent information",
    )
    parser.add_argument(
        "-j", "--json", action="store_true", help="Output in JSON format"
    )
    parser.add_argument("files", nargs="+", help="Files to analyze")
    args = parser.parse_args()

    # Build format_spec based on options
    fmt_options = []
    if args.hex:
        fmt_options.append("x")
    if args.verbose:
        fmt_options.append("v")
    if args.json:
        fmt_options.append("j")

    format_spec = ":".join(fmt_options)

    # Process each file
    for filepath in args.files:
        try:
            filemap = FileMap(filepath)
            print(f"{filemap:{format_spec}}")
        except Exception as e:
            print(f"Error processing {filepath}: {e}")


if __name__ == "__main__":
    main()



class Extent:
    # FIEMAP extent flag constants
    FIEMAP_EXTENT_LAST = 0x00000001
    FIEMAP_EXTENT_UNKNOWN = 0x00000002
    FIEMAP_EXTENT_DELALLOC = 0x00000004
    FIEMAP_EXTENT_ENCODED = 0x00000008
    FIEMAP_EXTENT_DATA_ENCRYPTED = 0x00000080
    FIEMAP_EXTENT_NOT_ALIGNED = 0x00000100
    FIEMAP_EXTENT_DATA_INLINE = 0x00000200
    FIEMAP_EXTENT_DATA_TAIL = 0x00000400
    FIEMAP_EXTENT_UNWRITTEN = 0x00000800
    FIEMAP_EXTENT_MERGED = 0x00001000
    FIEMAP_EXTENT_SHARED = 0x00002000

    def __init__(self, logical, physical, length, flags, device):
        self.logical = logical
        self.physical = physical
        self.length = length
        self.flags = flags
        self.device = device  # Device object

    def __repr__(self):
        return (
            f"<Extent(logical={self.logical}, physical={self.physical}, "
            f"length={self.length}, flags=0x{self.flags:X})>"
        )

    # Properties for flags
    @property
    def is_last(self):
        return bool(self.flags & self.FIEMAP_EXTENT_LAST)

    @property
    def is_unknown(self):
        return bool(self.flags & self.FIEMAP_EXTENT_UNKNOWN)

    @property
    def is_unwritten(self):
        return bool(self.flags & self.FIEMAP_EXTENT_UNWRITTEN)

    @property
    def is_shared(self):
        return bool(self.flags & self.FIEMAP_EXTENT_SHARED)

    # Add more properties as needed for other flags

    def __eq__(self, other):
        if not isinstance(other, Extent):
            return NotImplemented
        return (
            self.device == other.device
            and self.physical == other.physical
            and self.length == other.length
        )

    def __lt__(self, other):
        if not isinstance(other, Extent):
            return NotImplemented
        # Compare based on physical offset
        return self.physical < other.physical

    def __hash__(self):
        return hash((self.device, self.physical, self.length))

    def __format__(self, format_spec):
        fmt_options = set(format_spec.split(":"))
        use_hex = "x" in fmt_options
        verbose = "v" in fmt_options
        # json_output = 'j' in fmt_options  # Not implemented

        # Prepare the extent data
        logical = self.logical
        physical = self.physical
        length = self.length
        flags = self.flags

        # Format numbers in hexadecimal if requested
        if use_hex:
            logical_str = f"0x{logical:X}"
            physical_str = f"0x{physical:X}"
            length_str = f"0x{length:X}"
            flags_str = f"0x{flags:X}"
        else:
            logical_str = str(logical)
            physical_str = str(physical)
            length_str = str(length)
            flags_str = f"0x{flags:X}"

        # Interpret flags
        flag_descriptions = []
        if self.is_last:
            flag_descriptions.append("last")
        if self.is_unwritten:
            flag_descriptions.append("unwritten")
        if self.is_shared:
            flag_descriptions.append("shared")
        flags_readable = ",".join(flag_descriptions) if flag_descriptions else ""

        # Build output
        if verbose:
            output = (
                f"Extent(logical={logical_str}, physical={physical_str}, "
                f"length={length_str}, flags={flags_str} [{flags_readable}])"
            )
        else:
            output = f"Extent(logical={logical_str}, length={length_str})"

        return output

    def get_flag_descriptions(self):
        flag_descriptions = []
        if self.is_last:
            flag_descriptions.append("last")
        if self.is_unwritten:
            flag_descriptions.append("unwritten")
        if self.is_shared:
            flag_descriptions.append("shared")
        # Add other flag checks as needed
        return flag_descriptions



import os


class Device:
    def __init__(self):
        self.type = None
        self.id = None
        self.block_size = None
        self.source = None

    @classmethod
    def from_path(cls, path):
        instance = cls()

        try:
            # Get file system stats
            file_stats = os.stat(path)
            fs_stats = os.statvfs(path)

            # Populating fields
            instance.id = file_stats.st_dev
            instance.block_size = fs_stats.f_bsize

            # Determine device source using /proc/mounts
            instance.source = cls._get_device_source(path)

            # Determine device type
            instance.type = "block" if instance.source else "virtual"

        except FileNotFoundError:
            raise ValueError(f"Path {path} does not exist.")

        return instance

    @staticmethod
    def _get_device_source(path):
        # We'll store matching mount points and their corresponding devices
        matching_mounts = []

        # Read /proc/mounts to get all mounted filesystems
        with open("/proc/mounts", "r") as mounts_file:
            for line in mounts_file:
                parts = line.split()
                source_device = parts[0]  # The source device (e.g., /dev/sda1)
                mount_point = parts[1]  # The mount point (e.g., /)

                # Check if the given path starts with this mount point
                if path.startswith(mount_point):
                    matching_mounts.append((mount_point, source_device))

        if not matching_mounts:
            return None

        # Sort by the length of the mount point (longest first to ensure deepest match)
        matching_mounts.sort(key=lambda x: len(x[0]), reverse=True)

        # Return the source device for the deepest mount point
        return matching_mounts[0][1]

    def __repr__(self):
        return (
            f"<Device(type={self.type}, id={self.id}, block_size={self.block_size}, "
            f"source={self.source})>"
        )

    def __eq__(self, other):
        return isinstance(other, Device) and self.id == other.id



import json
import os

from . import fie
from .device import Device
from .extent import Extent


class FileMap:
    def __init__(self, path):
        self.path = path
        self.device = None
        self.inode = None
        self.mtime = None
        self.extents = []
        self.update()

    def update(self):
        # Get file stats
        try:
            file_stats = os.stat(self.path)
        except FileNotFoundError:
            raise ValueError(f"Path {self.path} does not exist.")

        self.device = Device.from_path(self.path)
        self.inode = file_stats.st_ino
        self.mtime = file_stats.st_mtime

        # Get extents
        fd = os.open(self.path, os.O_RDONLY)
        try:
            raw_extents = fie.get_extents(fd)
        finally:
            os.close(fd)

        # Update extents list
        self.extents = []
        for raw_extent in raw_extents:
            extent = Extent(
                logical=raw_extent["logical"],
                physical=raw_extent["physical"],
                length=raw_extent["length"],
                flags=raw_extent["flags"],
                device=self.device,
            )
            self.extents.append(extent)

    def check_stale(self):
        # Check if the file has changed since the last update
        try:
            file_stats = os.stat(self.path)
        except FileNotFoundError:
            return True  # File no longer exists

        if (
            self.device.id != file_stats.st_dev
            or self.inode != file_stats.st_ino
            or self.mtime != file_stats.st_mtime
        ):

            return True
        return False

    def __eq__(self, other):
        if not isinstance(other, FileMap):
            return NotImplemented
        return (
            self.device == other.device
            and self.inode == other.inode
            and self.mtime == other.mtime
        )

    def __iter__(self):
        return iter(self.extents)

    def __repr__(self):
        return f"<FileMap(path={self.path}, extents={len(self.extents)})>"

    def __format__(self, format_spec):
        fmt_options = set(format_spec.split(":"))
        verbose = "v" in fmt_options
        json_output = "j" in fmt_options

        if json_output:
            # Build JSON representation
            extents_data = [
                {
                    "logical": extent.logical,
                    "physical": extent.physical,
                    "length": extent.length,
                    "flags": extent.flags,
                    "flags_readable": extent.get_flag_descriptions(),
                }
                for extent in self.extents
            ]

            data = {
                "path": self.path,
                "device": {
                    "type": self.device.type,
                    "id": self.device.id,
                    "block_size": self.device.block_size,
                    "source": self.device.source,
                },
                "inode": self.inode,
                "mtime": self.mtime,
                "extents": extents_data,
            }
            return json.dumps(data, indent=2)

        else:
            # Build text output
            output = []
            output.append(f"File: {self.path}")
            output.append(f"Device: {self.device}")
            output.append(f"Inode: {self.inode}")
            output.append(f"Modification Time: {self.mtime}")
            output.append(f"Number of Extents: {len(self.extents)}")

            if verbose:
                for idx, extent in enumerate(self.extents):
                    # Remove 'j' option when formatting extents
                    extent_format_spec = ":".join(fmt_options - {"j"})
                    extent_str = format(extent, extent_format_spec)
                    output.append(f"  {idx}: {extent_str}")
            return "\n".join(output)
